# Copyright (c) LangChain, Inc.
# Licensed under the MIT License.

import os
import typing
import json
import pprint
import traceback

from langchain.output_parsers import StructuredOutputParser, ResponseSchema
from langchain.prompts import PromptTemplate, ChatPromptTemplate, HumanMessagePromptTemplate
from langchain_google_genai import ChatGoogleGenerativeAI
from langchain.output_parsers import CommaSeparatedListOutputParser
from langchain.schema import Document
from langchain_experimental.data_anonymizer import PresidioReversibleAnonymizer

from core.prompts import prompt_templates
from handlers.output_generator import generate_output

from pydantic import ValidationError

from dotenv import load_dotenv
load_dotenv()

llm = ChatGoogleGenerativeAI(model="gemini-pro", temperature=0.1)    

class LangchainFramework:

    def __init__(self):
        pass

    def language_detection(self, input_text: str) -> typing.Dict[str, typing.Any]:
        
        language_detection_struc: typing.Dict[str, typing.Any] = {}  # deep_linguistic_struc

        prompt = PromptTemplate.from_template(prompt_templates.language_detection_template)

        chain = prompt | llm

        context = input_text
        result = chain.invoke({"context": context})
        language_detection_struc['input_text'] = str(input_text)
        language_detection_struc['detected_language'] = str(result.content)

        return language_detection_struc  
    
    def sentiment_analysis(self, input_text: str) -> typing.Dict[str, typing.Any]:
        
        response_schemas = [
            ResponseSchema(name="input_text", description="add the input text here",type="string"),
            ResponseSchema(name="document_sentiment", description="Analyze the sentiment of the following text as positive, neutral, negative and mixed and provide respective confidence_scores",type="string"),
            ResponseSchema(name="positive", description="confidence scores for positive with precision 2", type="float"),
            ResponseSchema(name="negative", description="confidence scores for negative with precision 2", type="float"),
            ResponseSchema(name="neutral", description="confidence scores for neutral with precision 2", type="float"),
            ResponseSchema(name="mixed", description="confidence scores for mixed with precision 2", type="float")
        ]

        output_parser = StructuredOutputParser.from_response_schemas(response_schemas)
        format_instructions = output_parser.get_format_instructions()
        
        prompt = ChatPromptTemplate(
            messages=[
                HumanMessagePromptTemplate.from_template(prompt_templates.sentiment_analysis_template)
            ],
            input_variables=["question"],
            partial_variables={"format_instructions": format_instructions}
        )
        _input = prompt.format_prompt(question=input_text)
        chain = prompt | llm

        output  = chain.invoke({"question": _input.to_messages()})

        return output_parser.parse(output.content)
    
    def keyphrase_extraction(self, input_text: str) -> typing.Dict[str, typing.Any]:
        
        keyphrase_extraction_struc: typing.Dict[str, typing.Any] = {}  # deep_linguistic_struc

        output_parser = CommaSeparatedListOutputParser()

        format_instructions = output_parser.get_format_instructions()
        prompt = PromptTemplate(
            template=prompt_templates.keyphrase_extraction_template,
            input_variables=["context"],
            partial_variables={"format_instructions": format_instructions}
        )

        _input = prompt.format(context=input_text)
        chain = prompt | llm

        output  = chain.invoke({"context": _input})
        phrases = output_parser.parse(output.content)

        keyphrase_extraction_struc['mainPhrases'] = phrases

        return keyphrase_extraction_struc
    
    def pii_anonymization(self, document_content: str) -> typing.Dict[str, typing.Any]:
        
        pii_recognition_struc: typing.Dict[str, typing.Any] = {}  # deep_linguistic_struc
        pii_line_struc: typing.Dict[str, typing.Any] = {}  #  paragraphs_line_struc

        documents = [Document(page_content=document_content)]

        anonymizer = PresidioReversibleAnonymizer(
            add_default_faker_operators=False,
        )

        # print_colored_pii(anonymizer.anonymize(document_content))
        anonymizer.anonymize(document_content)

        pii_recognition_struc['pii_anonymizer_doc_before'] = str(anonymizer.anonymize(document_content))
        pii_recognition_struc['pii_deanonymizer_mapping_before'] = anonymizer.deanonymizer_mapping

        anonymizer = PresidioReversibleAnonymizer(
            add_default_faker_operators=True,
            # Faker seed is used here to make sure the same fake data is generated for the test purposes
            # In production, it is recommended to remove the faker_seed parameter (it will default to None)
            # faker_seed=42,
        )

        pii_recognition_struc['pii_anonymizer_doc_after'] = str(anonymizer.anonymize(document_content))
        pii_recognition_struc['pii_deanonymizer_mapping_after'] = anonymizer.deanonymizer_mapping

        return pii_recognition_struc




chat_words_str = """
AFAIK=As Far As I Know
AFK=Away From Keyboard
ASAP=As Soon As Possible
ATK=At The Keyboard
ATM=At The Moment
A3=Anytime, Anywhere, Anyplace
BAK=Back At Keyboard
BBL=Be Back Later
BBS=Be Back Soon
BFN=Bye For Now
B4N=Bye For Now
BRB=Be Right Back
BRT=Be Right There
BTW=By The Way
B4=Before
B4N=Bye For Now
CU=See You
CUL8R=See You Later
CYA=See You
FAQ=Frequently Asked Questions
FC=Fingers Crossed
FWIW=For What It's Worth
FYI=For Your Information
GAL=Get A Life
GG=Good Game
GN=Good Night
GMTA=Great Minds Think Alike
GR8=Great!
G9=Genius
IC=I See
ICQ=I Seek you (also a chat program)
ILU=ILU: I Love You
IMHO=In My Honest/Humble Opinion
IMO=In My Opinion
IOW=In Other Words
IRL=In Real Life
KISS=Keep It Simple, Stupid
LDR=Long Distance Relationship
LMAO=Laugh My A.. Off
LOL=Laughing Out Loud
LTNS=Long Time No See
L8R=Later
MTE=My Thoughts Exactly
M8=Mate
NRN=No Reply Necessary
OIC=Oh I See
PITA=Pain In The A..
PRT=Party
PRW=Parents Are Watching
ROFL=Rolling On The Floor Laughing
ROFLOL=Rolling On The Floor Laughing Out Loud
ROTFLMAO=Rolling On The Floor Laughing My A.. Off
SK8=Skate
STATS=Your sex and age
ASL=Age, Sex, Location
THX=Thank You
TTFN=Ta-Ta For Now!
TTYL=Talk To You Later
U=You
U2=You Too
U4E=Yours For Ever
WB=Welcome Back
WTF=What The F...
WTG=Way To Go!
WUF=Where Are You From?
W8=Wait...
7K=Sick:-D Laugher
"""


import logging
import traceback

from core.config import Settings
from handlers.custom_exceptions import CustomJSONError
from handlers.output_generator import generate_output
from core.constants import constants, preprocess_operations

from musaddiquehussainlabs.text_preprocessing.operations import operations_dict

logger = logging.getLogger(__name__)
if not logger.handlers:
    logger.setLevel(logging.DEBUG)

    file_handler = logging.FileHandler(Settings.LOG_FILE_PATH)
    file_handler.setLevel(logging.DEBUG)

    formatter = logging.Formatter(Settings.LOG_FILE_FORMAT)
    file_handler.setFormatter(formatter)

    logger.addHandler(file_handler)

# Please uncomment/comment below line if you wish to stop/start logging
# logger.disabled = True

def preprocess_text(text: str, operations=None) -> str:
    """ Preprocess an input text by executing a series of preprocessing functions specified in functions list """

    try:

        if text is None:
                raise CustomJSONError(constants.ERROR_CODE_500, constants.VALIDATE_INPUT_TEXT_NONE)
        
        if not isinstance(text, str):
                raise CustomJSONError(constants.ERROR_CODE_500, constants.VALIDATE_INPUT_TEXT_STRING)

        if not text.strip():
                raise CustomJSONError(constants.ERROR_CODE_500, constants.VALIDATE_INPUT_TEXT_EMPTY)
        
        default_operations = [
            preprocess_operations.to_lower,
            preprocess_operations.remove_html,
            preprocess_operations.remove_url,
            preprocess_operations.remove_email,
            preprocess_operations.remove_phone_number,
            preprocess_operations.remove_itemized_bullet_and_numbering,
            preprocess_operations.remove_punctuation,
            preprocess_operations.remove_special_character,
            preprocess_operations.remove_whitespace,
            preprocess_operations.normalize_unicode,
            preprocess_operations.expand_contraction,
            preprocess_operations.convert_emoticons_to_words,
            preprocess_operations.convert_emojis_to_words,
            preprocess_operations.chat_words_conversion,
            preprocess_operations.remove_stopword,
            preprocess_operations.remove_freqwords,
            preprocess_operations.remove_rarewords
        ]

        # If operations are not supplied, use default operations
        if operations is None:
            operations = default_operations

        # Validate text input
        if not isinstance(text, str):
            raise CustomJSONError(constants.ERROR_CODE_500, constants.VALIDATE_INPUT_TEXT_STRING)

        if operations is None or not isinstance(operations, list) or not operations:
            raise CustomJSONError(constants.ERROR_CODE_500, constants.TEXTPREPROCESSING_OPERATIONS_EMPTY_NONE)

        for operation_name in operations:
            if operation_name not in operations_dict:
                raise CustomJSONError(constants.ERROR_CODE_500, constants.TEXTPREPROCESSING_INVALID_OPERATIONS.format(operation_name))

        processed_text = text
        for operation_name in operations:
            operation_func = operations_dict.get(operation_name)
            if operation_func:
                processed_text = operation_func(processed_text)

                # Ensure the processed_text is a string after each operation
                if not isinstance(processed_text, str):
                    processed_text = ' '.join(processed_text)

            else:
                raise ValueError(constants.TEXTPREPROCESSING_OPERATIONS_NOT_CALLABLE.format(operation_name))

        return generate_output(True, data=processed_text)
    
    except CustomJSONError as e:
        logger.error(f"Custom JSON Error: {e.to_json()}")
        return e.to_json()

    except Exception as e:
        error_message = traceback.format_exc()
        logger.error(f"Unexpected error: {error_message}")
        return generate_output(False, error_code=500, message=constants.UNEXPECTED_ERROR)
